import java.util.LinkedList;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.lang.StringBuilder;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.tree.ParseTreeProperty;
import org.antlr.v4.runtime.tree.TerminalNode;

public class TypeChecker extends GrammarBaseVisitor<Type> {
  /** The scope containing the symbol table and the nested scopes. */
  public Scope scope;

  final String returnTypeSymbol = "$RETURNTYPE";

  @Override
  public Type visitProgram(GrammarParser.ProgramContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitDef(GrammarParser.DefContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitDecl(GrammarParser.DeclContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitAssignment(GrammarParser.AssignmentContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitWhile(GrammarParser.WhileContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitIfThenElse(GrammarParser.IfThenElseContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitIfThen(GrammarParser.IfThenContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitReturn(GrammarParser.ReturnContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitExprStmt(GrammarParser.ExprStmtContext ctx) {
    return Type.VOID;
  }

  @Override
  public Type visitEmpty(GrammarParser.EmptyContext ctx) {
    // nothing to do for empty statements
    return Type.VOID;
  }

  @Override
  public Type visitCompound(GrammarParser.CompoundContext ctx) {
    return Type.VOID;
  }

  /* types */

  @Override
  public Type visitIntType(GrammarParser.IntTypeContext ctx) {
    return Type.INT;
  }

  @Override
  public Type visitBoolType(GrammarParser.BoolTypeContext ctx) {
    return Type.BOOL;
  }

  @Override
  public Type visitStringType(GrammarParser.StringTypeContext ctx) {
    return Type.STRING;
  }

  @Override
  public Type visitFunType(GrammarParser.FunTypeContext ctx) {
    return new Type.FunctionType(null, null);  // TODO replace null
  }

  
  /* expressions */

  @Override
  public Type visitCall(GrammarParser.CallContext ctx) {
    return null;  // TODO replace with return type
  }

  @Override
  public Type visitNegate(GrammarParser.NegateContext ctx) {
    return Type.INT;
  }

  @Override
  public Type visitNot(GrammarParser.NotContext ctx) {
    return Type.BOOL;
  }

  @Override
  public Type visitMultDiv(GrammarParser.MultDivContext ctx) {
    return Type.INT;
  }

  @Override
  public Type visitAddSub(GrammarParser.AddSubContext ctx) {
    return Type.INT;
  }

  @Override
  public Type visitRelational(GrammarParser.RelationalContext ctx) {
    return Type.BOOL;
  }

  @Override
  public Type visitEqNeq(GrammarParser.EqNeqContext ctx) {
    return Type.BOOL;
  }

  @Override
  public Type visitAndOr(GrammarParser.AndOrContext ctx) {
    return Type.BOOL;
  }

  @Override
  public Type visitVar(GrammarParser.VarContext ctx) {
    return null;  // TODO get type of the symbol
  }
    
  @Override
  public Type visitNum(GrammarParser.NumContext ctx) {
    info(ctx, String.format("number constants always have type int"));
    return Type.INT;
  }
    
  @Override
  public Type visitStringLiteral(GrammarParser.StringLiteralContext ctx) {
    info(ctx, String.format("string literals always have type string"));
    return Type.STRING;
  }
    
  @Override
  public Type visitParens(GrammarParser.ParensContext ctx) {
    return visit(ctx.expr());
  }

  /* name */

  @Override
  public Type visitSimple(GrammarParser.SimpleContext ctx) {
    return null;  // TODO return type of the name
  }

  /** Print informational messages at a specific file, line, and column. */
  public static void info(ParserRuleContext ctx, String msg) {
    Token firstToken = ctx.getStart();
    String file = firstToken.getTokenSource().getSourceName();
    int line = firstToken.getLine();
    int col = firstToken.getCharPositionInLine() + 1;
    System.err.println(String.format("[INFO] %s:%d:%d %s", file, line, col, msg));
  }
}
